@@ -221,6 +221,8 @@ static gboolean     gtk_file_folder_gnome_vfs_list_children (GtkFileFolder
 							     GSList            **children,
 							     GError            **error);
 
+static gboolean     gtk_file_folder_gnome_vfs_is_finished_loading (GtkFileFolder *folder);
+
 static GtkFileInfo *           info_from_vfs_info (const gchar      *uri,
 						   GnomeVFSFileInfo *vfs_info,
 						   GtkFileInfoType   types);
@@ -375,14 +377,14 @@ gtk_file_system_gnome_vfs_init (GtkFileSystemGnomeVFS *system_vfs)
 #endif
 
   system_vfs->volume_monitor = gnome_vfs_volume_monitor_ref (gnome_vfs_get_volume_monitor ());
-  g_signal_connect (system_vfs->volume_monitor, "volume-mounted",
-		    G_CALLBACK (volume_mount_unmount_cb), system_vfs);
-  g_signal_connect (system_vfs->volume_monitor, "volume-unmounted",
-		    G_CALLBACK (volume_mount_unmount_cb), system_vfs);
-  g_signal_connect (system_vfs->volume_monitor, "drive-connected",
-		    G_CALLBACK (drive_connect_disconnect_cb), system_vfs);
-  g_signal_connect (system_vfs->volume_monitor, "drive-disconnected",
-		    G_CALLBACK (drive_connect_disconnect_cb), system_vfs);
+  g_signal_connect_object (system_vfs->volume_monitor, "volume-mounted",
+			   G_CALLBACK (volume_mount_unmount_cb), system_vfs, 0);
+  g_signal_connect_object (system_vfs->volume_monitor, "volume-unmounted",
+			   G_CALLBACK (volume_mount_unmount_cb), system_vfs, 0);
+  g_signal_connect_object (system_vfs->volume_monitor, "drive-connected",
+			   G_CALLBACK (drive_connect_disconnect_cb), system_vfs, 0);
+  g_signal_connect_object (system_vfs->volume_monitor, "drive-disconnected",
+			   G_CALLBACK (drive_connect_disconnect_cb), system_vfs, 0);
 
 }
 
@@ -463,7 +465,8 @@ gtk_file_system_gnome_vfs_list_volumes (GtkFileSystem *file_system)
 
   /* Add root, first in the list */
   volume = gnome_vfs_volume_monitor_get_volume_for_path (system_vfs->volume_monitor, "/");
-  result = g_slist_prepend (result, volume);
+  if (volume)
+    result = g_slist_prepend (result, volume);
   
   return result;
 }
@@ -525,12 +528,14 @@ load_dir (GtkFileFolderGnomeVFS *folder_vfs)
 
   g_hash_table_foreach_remove (folder_vfs->children,
 			       remove_all, NULL);
+  gnome_authentication_manager_push_async ();
   gnome_vfs_async_load_directory (&folder_vfs->async_handle,
 				  folder_vfs->uri,
 				  get_options (folder_vfs->types),
 				  ITEMS_PER_NOTIFICATION,
 				  GNOME_VFS_PRIORITY_DEFAULT,
 				  directory_load_callback, folder_vfs);
+  gnome_authentication_manager_pop_async ();
 }
 
 static char *
@@ -562,6 +567,8 @@ gtk_file_system_gnome_vfs_get_folder (GtkFileSystem     *file_system,
 				      GError           **error)
 {
   char *uri;
+  GnomeVFSURI *vfs_uri;
+  gboolean exists;
   GtkFilePath *parent_path;
   GtkFileSystemGnomeVFS *system_vfs = GTK_FILE_SYSTEM_GNOME_VFS (file_system);
   GtkFileFolderGnomeVFS *folder_vfs;
@@ -583,12 +590,34 @@ gtk_file_system_gnome_vfs_get_folder (GtkFileSystem     *file_system,
       return NULL;
     }
 
+  vfs_uri = gnome_vfs_uri_new (uri);
+  if (!vfs_uri)
+    {
+      set_vfs_error (GNOME_VFS_ERROR_INVALID_URI, uri, error);
+      gtk_file_path_free (parent_path);
+      g_free (uri);
+      return NULL;
+    }
+
+  exists = gnome_vfs_uri_exists (vfs_uri);
+  gnome_vfs_uri_unref (vfs_uri);
+
+  if (!exists)
+    {
+      set_vfs_error (GNOME_VFS_ERROR_NOT_FOUND, uri, error);
+      gtk_file_path_free (parent_path);
+      g_free (uri);
+      return NULL;
+    }
+
   folder_vfs = g_object_new (GTK_TYPE_FILE_FOLDER_GNOME_VFS, NULL);
 
+  gnome_authentication_manager_push_sync ();
   result = gnome_vfs_monitor_add (&monitor,
 				  uri,
 				  GNOME_VFS_MONITOR_DIRECTORY,
 				  monitor_callback, folder_vfs);
+  gnome_authentication_manager_pop_sync ();
   if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_NOT_SUPPORTED)
     {
       g_free (uri);
@@ -627,9 +656,11 @@ gtk_file_system_gnome_vfs_get_folder (GtkFileSystem     *file_system,
 	  FolderChild *child;
 
 	  vfs_info = gnome_vfs_file_info_new ();
+	  gnome_authentication_manager_push_sync ();
 	  result = gnome_vfs_get_file_info (folder_vfs->uri,
 					    vfs_info,
 					    get_options (parent_folder->types));
+	  gnome_authentication_manager_pop_sync ();
 	  if (result != GNOME_VFS_OK)
 	    {
 	      gnome_vfs_file_info_unref (vfs_info);
@@ -661,10 +692,12 @@ gtk_file_system_gnome_vfs_create_folder (GtkFileSystem     *file_system,
   const gchar *uri = gtk_file_path_get_string (path);
   GnomeVFSResult result;
 
+  gnome_authentication_manager_push_sync ();
   result = gnome_vfs_make_directory (uri,
 				     GNOME_VFS_PERM_USER_ALL |
 				     GNOME_VFS_PERM_GROUP_ALL |
 				     GNOME_VFS_PERM_OTHER_READ);
+  gnome_authentication_manager_pop_sync ();
 
   if (result != GNOME_VFS_OK)
     {
@@ -764,7 +797,9 @@ gtk_file_system_gnome_vfs_volume_mount (GtkFileSystem        *file_system,
 
       closure.system_vfs = system_vfs;
       closure.loop = g_main_loop_new (NULL, FALSE);
+      gnome_authentication_manager_push_sync ();
       gnome_vfs_drive_mount (GNOME_VFS_DRIVE (volume), volume_mount_cb, &closure);
+      gnome_authentication_manager_pop_sync ();
 
       GDK_THREADS_LEAVE ();
       g_main_loop_run (closure.loop);
@@ -928,7 +963,7 @@ gtk_file_system_gnome_vfs_volume_render_icon (GtkFileSystem        *file_system,
     {
       uri = gnome_vfs_volume_get_activation_uri (GNOME_VFS_VOLUME (volume));
       if (strcmp (uri, "file:///") == 0)
-	icon_name = g_strdup ("gnome-fs-blockdev");
+	icon_name = g_strdup ("gnome-dev-harddisk");
       else if (strcmp (uri, system_vfs->desktop_uri) == 0)
 	icon_name = g_strdup ("gnome-fs-desktop");
       else if (strcmp (uri, system_vfs->home_uri) == 0)
@@ -1236,7 +1271,10 @@ static GtkFilePath *
 gtk_file_system_gnome_vfs_uri_to_path (GtkFileSystem *file_system,
 				       const gchar   *uri)
 {
-  return gtk_file_path_new_dup (uri);
+  gchar *canonical;
+
+  canonical = make_uri_canonical (uri);
+  return gtk_file_path_new_steal (canonical);
 }
 
 static GtkFilePath *
@@ -1245,7 +1283,14 @@ gtk_file_system_gnome_vfs_filename_to_path (GtkFileSystem *file_system,
 {
   gchar *uri = gnome_vfs_get_uri_from_local_path (filename);
   if (uri)
-    return gtk_file_path_new_steal (uri);
+    {
+      gchar *canonical;
+
+      canonical = make_uri_canonical (uri);
+      g_free (uri);
+
+      return gtk_file_path_new_steal (canonical);
+    }
   else
     return NULL;
 }
@@ -1290,7 +1335,9 @@ get_vfs_info (GtkFileSystem     *file_system,
     {
       info = gnome_vfs_file_info_new ();
       uri = gtk_file_path_get_string (path);
+      gnome_authentication_manager_push_sync ();
       gnome_vfs_get_file_info (uri, info, get_options (types));
+      gnome_authentication_manager_pop_sync ();
     }
   
   gtk_file_path_free (parent_path);
@@ -1748,6 +1795,7 @@ gtk_file_folder_gnome_vfs_iface_init (GtkFileFolderIface *iface)
 {
   iface->get_info = gtk_file_folder_gnome_vfs_get_info;
   iface->list_children = gtk_file_folder_gnome_vfs_list_children;
+  iface->is_finished_loading = gtk_file_folder_gnome_vfs_is_finished_loading;
 }
 
 static void
@@ -1782,9 +1830,10 @@ lookup_vfs_info_in_folder (GtkFileFolder     *folder,
 			   GError           **error)
 {
   GtkFileFolderGnomeVFS *folder_vfs = GTK_FILE_FOLDER_GNOME_VFS (folder);
-  const gchar *uri = gtk_file_path_get_string (path);
+  const gchar *uri;
   FolderChild *child;
 
+  uri = make_uri_canonical (gtk_file_path_get_string (path));
   child = g_hash_table_lookup (folder_vfs->children, uri);
   if (!child)
     {
@@ -1794,7 +1843,9 @@ lookup_vfs_info_in_folder (GtkFileFolder     *folder,
       GnomeVFSResult result;
       
       vfs_info = gnome_vfs_file_info_new ();
+      gnome_authentication_manager_push_sync ();
       result = gnome_vfs_get_file_info (uri, vfs_info, get_options (folder_vfs->types));
+      gnome_authentication_manager_pop_sync ();
       
       if (result != GNOME_VFS_OK)
 	set_vfs_error (result, uri, error);
@@ -1841,7 +1892,9 @@ gtk_file_folder_gnome_vfs_get_info (GtkFileFolder     *folder,
       gnome_vfs_uri_unref (vfs_uri);
 
       info = gnome_vfs_file_info_new ();
+      gnome_authentication_manager_push_sync ();
       result = gnome_vfs_get_file_info (folder_vfs->uri, info, get_options (GTK_FILE_INFO_ALL));
+      gnome_authentication_manager_pop_sync ();
       if (result != GNOME_VFS_OK)
 	{
 	  file_info = NULL;
@@ -1891,6 +1944,14 @@ gtk_file_folder_gnome_vfs_list_children (GtkFileFolder  *folder,
   return TRUE;
 }
 
+static gboolean
+gtk_file_folder_gnome_vfs_is_finished_loading (GtkFileFolder *folder)
+{
+  GtkFileFolderGnomeVFS *folder_vfs = GTK_FILE_FOLDER_GNOME_VFS (folder);
+
+  return (folder_vfs->async_handle == NULL);
+}
+
 
 static GnomeVFSFileInfoOptions
 get_options (GtkFileInfoType types)
@@ -2098,7 +2159,11 @@ directory_load_callback (GnomeVFSAsyncHandle *handle,
     }
 
   if (result != GNOME_VFS_OK)
-    folder_vfs->async_handle = NULL;
+    {
+      folder_vfs->async_handle = NULL;
+
+      g_signal_emit_by_name (folder_vfs, "finished-loading");
+    }
 }
 
 static void
@@ -2120,7 +2185,9 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
 	GnomeVFSFileInfo *vfs_info;
 
 	vfs_info = gnome_vfs_file_info_new ();
+	gnome_authentication_manager_push_sync ();
 	result = gnome_vfs_get_file_info (info_uri, vfs_info, get_options (folder_vfs->types));
+	gnome_authentication_manager_pop_sync ();
 
 	if (result == GNOME_VFS_OK)
 	  {
@@ -2313,6 +2380,15 @@ fs_module_init (GTypeModule    *module)
 				 GTK_TYPE_FILE_FOLDER,
 				 &file_folder_info);
   }
+
+  /* We ref the class so that the module won't get unloaded.  We need this
+   * because gnome_icon_lookup() will cause GnomeIconTheme to get registered,
+   * and we cannot unload that properly.  I'd prefer to do this via
+   * g_module_make_resident(), but here we don't have access to the GModule.
+   * See bug #139254 for reference. [federico]
+   */
+
+  g_type_class_ref (type_gtk_file_system_gnome_vfs);
 }
 
 void 
